/**
* @package scheduler404
* @copyright (C) 2004 - 2005 erreur404.org
* @license http://www.gnu.org/copyleft/gpl.html GNU/GPL
* MediaBox404 is Free Software
* @author vortex@erreur404.org
* @file scheduler404.c
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* header */
#include "scheduler404.h"

/* std */
#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/* mysql */
#include <mysql/mysql.h>

/* files */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/* sockets */
#include <netdb.h>
#include <sys/socket.h>
#include <resolv.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/* scheduler404 */
#include "xml_config.h"
#include "mysql_utils.h"
#include "log.h"
#include "util.h"

/* pthread */
#include <pthread.h>

#define BUFSIZE 4096


/* BEGIN : All following variables are not protected and are used in "main" and in thread. They are a good reason to "segfault" !! */
/* configuration structure */
mediabox404_config_t mediabox404_config;
int flag_doublon_jingle_prg = 0;
int StreamType = MAIN_STREAM; /* 0:MAIN,1:PL,2:REDIF,3:LIVE,4:RELAY */
int CurrentPlaylist = 0;
int nb_control_play = 0;
time_t EndRelay = 0;
/* END of dangerous variables */

/* update function */
int scheduler_check_playlist();

/* thread */
pthread_t synchronize_pthread;
pthread_t listen_stream404_pthread;
void *scheduler_listen_stream404(void *unused);

static int stream404_current_sd;			/* Current stream404 socket descriptor */
static pthread_mutex_t mutex_stream404_sd = PTHREAD_MUTEX_INITIALIZER;

/** 
 * @fn void *scheduler_synchronize(void *unused)
 * @param unused unused.
 * @return nothing.
 * @brief Listen connections and commands from core404 and control stream404
 */
void *scheduler_synchronize(void *unused) {
  //struct hostent* host; /* host structure */
  MYSQL mysql; /* MySQL main handler */
  static int core404_server_sd = 0;		/* core404's server socket descriptor */
  static struct sockaddr_in core404_addr;	/* address structure used to bind socket */
  char buffer[BUFSIZ];

  mediabox404_log_debug("THREAD: thread started.");

  /* MySQL init & connect */
  mysql_init(&mysql);
  mysql_options(&mysql, MYSQL_READ_DEFAULT_GROUP, mediabox404_config.sql.mysql_options);
  mediabox404_log_debug("THREAD: MySQL initialized.");

  mediabox404_log_debug("THREAD: Connecting to database...");
  if (!mediabox404_config.sql.usocket) {
    /* connect using host parameter */
    if ( !mysql_real_connect(&mysql, mediabox404_config.sql.hostname, mediabox404_config.sql.user, mediabox404_config.sql.password, mediabox404_config.sql.database, 0, NULL,0)) {
      mediabox404_log ("%s : Cannot connect port to sql database ! hostname:%s user:%s password:%s database:%s %d",
          mysql_error(&mysql),
          mediabox404_config.sql.hostname, 
          mediabox404_config.sql.user, 
          mediabox404_config.sql.password, 
          mediabox404_config.sql.database,
          mediabox404_config.sql.port);
      exit(EXIT_FAILURE);
    }
  } else {
    /* connect using socket parameter */
      if (!mysql_real_connect(&mysql, NULL, mediabox404_config.sql.user, 
          mediabox404_config.sql.password, 
          mediabox404_config.sql.database,
          mediabox404_config.sql.port, 
          mediabox404_config.sql.usocket, 0)) {

      mediabox404_log ("%s : Cannot connect usocket to sql database ! hostname:%s user:%s password:%s database:%s %d",
      mysql_error(&mysql),
      mediabox404_config.sql.hostname, 
          mediabox404_config.sql.user, 
          mediabox404_config.sql.password, 
          mediabox404_config.sql.database,
          mediabox404_config.sql.usocket);
      exit(EXIT_FAILURE);
    }
  }
  mediabox404_log_debug("THREAD: Database connected.");

  mediabox404_log_debug("THREAD: Creating core404's socket");
  core404_server_sd = socket(PF_INET, SOCK_STREAM, 0); /* create socket */
  if ( core404_server_sd < 0 ) {
    mediabox404_log("THREAD: Unable to create socket, exiting.");
    exit(EXIT_FAILURE);
  }
  mediabox404_log_debug("THREAD: stream404's socket created.");
  memset(&core404_addr, 0, sizeof(core404_addr));	/* create & zero struct */
  core404_addr.sin_family = AF_INET;		/* select internet protocol */
  core404_addr.sin_port = htons(4200);			/* set the port # */
  core404_addr.sin_addr.s_addr = INADDR_ANY;		/* any interface */
  // core404_addr.sin_addr.s_addr = INADDR_LOOPBACK /* = 127.0.0.1 */
  //*(long*)host->h_addr_list[0]; /* set the addr */

  mediabox404_log_debug("THREAD: Binding core404's socket...");
  if ( -1 == bind(core404_server_sd, &core404_addr, sizeof(core404_addr)) ) { /* connect! */
    mediabox404_log_debug("THREAD: Unable to bind core404's socket...");
    exit(EXIT_FAILURE);
  }
  mediabox404_log_debug("THREAD: core404's socket binded.");

  listen(core404_server_sd, 2);				/* make into listener with 2 slots */

  stream404_current_sd = 0;

  mediabox404_log_debug("THREAD: First check of scheduled playlist...");
  if ( scheduler_check_playlist( &mysql ) == 1 ) {
    mediabox404_log_debug("THREAD: First scheduled playlist check done.");
  }

  /* Launch the thread that will accept connections from stream404 */
  pthread_create( &listen_stream404_pthread, NULL, scheduler_listen_stream404, NULL);

  while (42) {
    static int core404_client_sd;		/* core404's client socket descriptor */
    static FILE* core404_fp;			/* core404's client socket descriptor converted into stream */

    core404_client_sd = accept(core404_server_sd, 0, 0);    /* accept connection */
    mediabox404_log_debug("THREAD: New core404 client accepted...");

    core404_fp = fdopen(core404_client_sd, "r+");    /* convert into FILE* */
    /*--- Process client's requests ---*/
    while (fgets(buffer, sizeof(buffer), core404_fp) != 0) {
      /* proc client's requests */
      if ( 0 == strcmp(buffer, "exit\n") ) {
        mediabox404_log_debug("THREAD: Exit command have been received...");
        exit(EXIT_SUCCESS);
      } else if ( 0 == strcmp(buffer, "sync\n") ) {
        mediabox404_log_debug("THREAD: Sync request received...");
        /* Update scheduled playlist */
        mediabox404_log_debug("THREAD: Checking playlist...");
        if ( scheduler_check_playlist( &mysql ) == 1 ) {
          mediabox404_log_debug("THREAD: Playlist checked.");
        }

        /* Send synchronization request to the stream404 client */
        pthread_mutex_lock(&mutex_stream404_sd);
        if (stream404_current_sd) {
          write(stream404_current_sd, "sync\n", 5);
        }
        pthread_mutex_unlock(&mutex_stream404_sd);
      } else if ( 0 == strcmp(buffer, "next\n") ) {
        /* Send next request directly to the stream404 client */
        pthread_mutex_lock(&mutex_stream404_sd);
        if (stream404_current_sd) {
          write(stream404_current_sd, "next\n", 5);
        }
        pthread_mutex_unlock(&mutex_stream404_sd);
      }  else {
        mediabox404_log_debug("THREAD: Unknown command received : \"%s\" !", buffer);
      }
    }
    fclose(core404_fp);         /* close the client's channel */
  }
}

/** 
 * @fn void *scheduler_listen_stream404(void *unused)
 * @param unused unused.
 * @return nothing.
 * @brief Accepts connections from stream404 
 */
void *scheduler_listen_stream404(void *unused) {
	static int stream404_server_sd;		/* stream404's socket descriptor */
	static struct sockaddr_in addr;	/* address structure used to bind socket */

	mediabox404_log_debug("THREAD: Creating stream404's socket");
	stream404_server_sd = socket(PF_INET, SOCK_STREAM, 0); /* create socket */
	if ( stream404_server_sd < 0 ) {
		mediabox404_log("THREAD: Unable to create socket, exiting.");
		exit(EXIT_FAILURE);
	}
	mediabox404_log_debug("THREAD: stream404's socket created.");
	memset(&addr, 0, sizeof(addr));	/* create & zero struct */
	addr.sin_family = AF_INET;		/* select internet protocol */
	addr.sin_port = htons(4201);			/* set the port # */
	addr.sin_addr.s_addr = INADDR_ANY;		/* any interface */
	//*(long*)host->h_addr_list[0]; /* set the addr */

	mediabox404_log_debug("THREAD: Binding stream404's socket...");
	if ( -1 == bind(stream404_server_sd, &addr, sizeof(addr)) ) { /* connect! */
		mediabox404_log_debug("THREAD: Unable to bind stream404's socket...");
		exit(EXIT_FAILURE);
	}
	mediabox404_log_debug("THREAD: stream404's socket binded.");

	listen(stream404_server_sd, 2);				/* make into listener with 2 slots */
	mediabox404_log_debug("THREAD: Listening stream404's socket...");

	while (42) {
		int stream404_new_sd;
		stream404_new_sd = accept(stream404_server_sd, 0, 0);    /* accept connection */
		mediabox404_log_debug("THREAD: New stream404 client accepted...");

		if ( stream404_current_sd != stream404_new_sd ) {
			pthread_mutex_lock(&mutex_stream404_sd);
			if (stream404_current_sd) {
				close(stream404_current_sd);     /* close last client's channel */
			}
			stream404_current_sd = stream404_new_sd;    /* convert into stream */
			pthread_mutex_unlock(&mutex_stream404_sd);
		}
	}
}


/** 
 * @fn int scheduler_check_playlist( MYSQL* mysql )
 * @param mysql used to handle a configured MySQL connection..
 * @return update code (unused a this time, always 1)
 * @brief Stop "Live" or "Relay" if curent stream is "Live" or "Relay"
 */
int scheduler_check_playlist( MYSQL* mysql ) {
	/* MySQL */
	MYSQL_RES *res;
	MYSQL_ROW row;
	int query_ok;

	/* ??? */
	char CODE[10];
	char ARG[8][512];
// 
	unsigned long *lengths;
	unsigned int StackPrgID, UserPrgID, DataPrgID, LIVE_Bitrate, LIVE_Frequence, LIVE_NbCanal, LIVE_NbReconnect;
	int wait=0;
	char query_prgstack[255];

	mediabox404_log_debug("CHECKPL: Start checking...");
	sprintf(query_prgstack, "SELECT * FROM %s ORDER BY s_id ASC;", mediabox404_config.sql.progstack);
	mediabox404_log_debug("CHECKPL: Executing query : %s", query_prgstack);
	query_ok = mysql_real_query(mysql, query_prgstack, (unsigned int) strlen(query_prgstack));
	mediabox404_log_debug("CHECKPL: Query executed.");

	if (query_ok == 0) {
		res = mysql_store_result(mysql);
		mediabox404_log_debug("CHECKPL: Query result taken.");
		if (mysql_num_rows(res) > 0) {
			mediabox404_log_debug("CHECKPL: There is some rows in query.");
			while ( (row = mysql_fetch_row(res)) ) {
				mediabox404_log_debug("CHECKPL: 1 row taken...");
				lengths = mysql_fetch_lengths(res);
				sprintf(CODE,"%.*s", (int) lengths[2], row[2] ? row[2] : "NULL"); // Command Code LIVE
				StackPrgID = atoi(row[0]);

				//logfile("Pilot=%s\n",CODE);
				mediabox404_log_debug("CHECKPL: CODE : %s.", CODE);
				if (strcmp(CODE,"LIVE") == 0) {
					UserPrgID = atoi(row[1]);
					DataPrgID = atoi(row[4]);
					sprintf(ARG[0],"%.*s", (int) lengths[5], row[5] ? row[5] : "NULL"); // URL
					sprintf(ARG[1],"%.*s", (int) lengths[6], row[6] ? row[6] : "NULL"); // Format
					LIVE_Bitrate = atoi(row[7]);// Bitrate [16000,24000,32000,48000,56000,64000,80000,96000,112000,128000,160000,196000,256000]
					LIVE_Frequence = atoi(row[8]); // Frequency [16000,22050,24000,32000,44100,48000]
					LIVE_NbCanal = atoi(row[9]);// Channel [1,2] Mono/Stereo
					LIVE_NbReconnect = atoi(row[10]); // Max retries count
					//insert LIVE dans T_StreamStack;
					Insert_Log(mysql,&mediabox404_config,0,"LIVE",DataPrgID,"","","","","","","","");
					Insert_streamstack_live(mysql,&mediabox404_config,UserPrgID,DataPrgID,ARG[0],ARG[1],LIVE_Bitrate,LIVE_Frequence,LIVE_NbCanal,LIVE_NbReconnect,0);
					// Insert Jingle
					Insert_Jingle(mysql,&mediabox404_config);
				}

				if (strcmp(CODE,"RUN") == 0) {
					mediabox404_log_debug("End Relay manuel");
					CurrentPlaylist = 0;
					flag_doublon_jingle_prg = 0;
					StreamType = MAIN_STREAM;
					// copy pile t_tempstack --> stream , delete pile t_tempstack
					Copy_tempstack_streamstack(mysql,&mediabox404_config);
					do {
						Insert_Main_PL(mysql,&mediabox404_config,&nb_control_play);
					} while (nb_control_play != 0);
					// on arret le LIVE
					Stop_Current_Stream(mysql,&mediabox404_config);
					EndRelay = 0;
				}
				// Purge element admin ID
				Delete_ProgStackID (mysql,&mediabox404_config,StackPrgID);
				wait = 1;
			}
		} else {
			mediabox404_log_debug("CHECKPL: No row in query result!");
		}
		mysql_free_result(res);
	} else {
		mediabox404_log_debug("CHECKPL: Error while executing query!");
	}

	if (wait == 1) {
		mediabox404_log_debug("CHECKPL: Waiting for a core404 command %s", mediabox404_config.sql.mysql_options);
		wait = 0;
	}

	return 1;
}

int main(int argc, char *argv[])
{
	/* mysql */
	MYSQL mysql; /* MySQL main handler */
	int query_ok;

	int flag_doublon_promo = 0;
	int flag_doublon_jingle = 0;
//	int wait=0;

	unsigned int PlayTime;

	/* others */
	time_t timer = 0;
	struct tm *today;

	time_t NextAutoProgramme = 0;
	time_t NextScanProgramme = 0;
	time_t NextScanJingle = 0;
	time_t NextScanPromo = 0;

	char query_streamstack[255];


	int CurrentData = 0;

	int ScanPrg = 0; /* how many seconds to scan t_progstack stack */
	int flag_stop_relay = 0;

	unsigned int AutoProgDelay;
	unsigned int control;

	char MetaData[512];


	//    char CODE[10];
	char date_prg[32];

	int i;

	unsigned int data_id,prg_id,prg_type;
#define TIMINGVALUELEN 22
	int TimingValue[TIMINGVALUELEN] = {1, 2, 3, 4, 5, 6, 8, 9, 10, 12, 15, 16, 
		18, 20, 24, 30, 32, 36, 40, 45, 48, 60};



	/***********************/
	mediabox404_config.verbose = 1;

	if (argc != 2) {
		mediabox404_log("Define xml config file !!!");
		mediabox404_log("IE: mediabox404_radio_prg mediabox404_radio_prg.xml");
		mediabox404_log("Exit Programmation Mediabox404_audio");

		return EXIT_SUCCESS;
	}


	/* Parse the options in the config file, and the command line */
	mediabox404_xml_parse_config_file (&mediabox404_config, argv[1]);

	mysql_init(&mysql);
	mysql_options(&mysql, MYSQL_READ_DEFAULT_GROUP, mediabox404_config.sql.mysql_options);

	if (!mediabox404_config.sql.usocket) {
		/* connect using host parameter */
		if (!mysql_real_connect(&mysql, mediabox404_config.sql.hostname, 
					mediabox404_config.sql.user, 
					mediabox404_config.sql.password, 
					mediabox404_config.sql.database, 0, NULL,0)) {
			mediabox404_log ("%s : Cannot connect port to sql database ! hostname:%s user:%s password:%s database:%s %d",
					mysql_error(&mysql),
					mediabox404_config.sql.hostname, 
					mediabox404_config.sql.user, 
					mediabox404_config.sql.password, 
					mediabox404_config.sql.database,
					mediabox404_config.sql.port);
			exit(EXIT_FAILURE);
		}
	} else {
		/* connect using socket parameter */
		if (!mysql_real_connect(&mysql, NULL, mediabox404_config.sql.user, 
					mediabox404_config.sql.password, 
					mediabox404_config.sql.database,
					mediabox404_config.sql.port, 
					mediabox404_config.sql.usocket, 0)) {
			mediabox404_log ("%s : Cannot connect usocket to sql database ! hostname:%s user:%s password:%s database:%s %d",
					mysql_error(&mysql),
					mediabox404_config.sql.hostname, 
					mediabox404_config.sql.user, 
					mediabox404_config.sql.password, 
					mediabox404_config.sql.database,
					mediabox404_config.sql.usocket);
			exit(EXIT_FAILURE);
		}
	}

	mediabox404_log("Starting %s", 	mediabox404_config.sql.mysql_options);
	mediabox404_log_debug("Connected...");
	mediabox404_log_debug("Timming control...");

	/* find the timing value */
	control = 0;
	for (i = 0 ; i < TIMINGVALUELEN; i++) {
		if (TimingValue[i] == mediabox404_config.program_timing) {
			control++;
		}
	}

	if (control != 1) {
		mediabox404_log_debug("Program_timing error : %d not in {1,2,3,4,5,6,8,9,10,12,15,16,18,20,24,30,32,36,40,45,48,60} list",
				mediabox404_config.program_timing);
	}

	if ((mediabox404_config.jingle_timing % 
				mediabox404_config.program_timing) == 0) {
		control++;
	} else {
		mediabox404_log_debug("Jingle_timing error : %d not mod program_timing %d",
				mediabox404_config.jingle_timing,
				mediabox404_config.program_timing);
	}

	if ((mediabox404_config.promo_timing % 
				mediabox404_config.program_timing)==0) {
		control++;
	} else {
		mediabox404_log_debug("Promo_timing error : %d not mod program_timing %d",
				mediabox404_config.promo_timing,
				mediabox404_config.program_timing);
	}

	if (control != 3) {
		mediabox404_log("Timing control error...");
		mediabox404_log("Exit Programmation Mediabox404_audio");

		return EXIT_FAILURE;
	}

	mediabox404_log_debug("Timming control... ok");

	time(&timer);

	today = localtime(&timer);
	today->tm_sec = 0;
	today->tm_min = 0;
	NextScanProgramme = mktime(today);
	NextScanJingle = mktime(today);
	NextScanPromo = mktime(today);

	/* launch a jingle when starting */
	mediabox404_log_debug("Purge streamstack");
	Delete_StreamStack(&mysql, &mediabox404_config);
	Insert_Jingle(&mysql, &mediabox404_config);

	while (NextScanProgramme <= timer) {
		NextScanProgramme = NextScanProgramme + 
			(mediabox404_config.program_timing * 60);
	}

	while (NextScanJingle <= timer) {
		NextScanJingle = NextScanJingle + (mediabox404_config.jingle_timing * 60);
	}

	while (NextScanPromo <= timer) {
		NextScanPromo = NextScanPromo + (mediabox404_config.promo_timing * 60);
	}

	/***********************/
	pthread_create( &synchronize_pthread, NULL, scheduler_synchronize, NULL);

	/* scheduling main loop */
	while (42) {
		flag_stop_relay = 0;
		time( &timer );
		today = localtime( &timer );
		sprintf(date_prg, "%04d-%02d-%02d %02d:%02d:%02d",
				1900+today->tm_year, today->tm_mon+1,
				today->tm_mday, today->tm_hour,
				today->tm_min, today->tm_sec);

		/* "Promo" scheduling function */
		if (timer >= NextScanPromo) {
			/* it's time to schedule a promo... */
			if (flag_doublon_promo == 0) {
				flag_doublon_promo = 1;
				if (StreamType == MAIN_STREAM) {
					Insert_Promo(&mysql, &mediabox404_config);
				}
				today = localtime(&timer);
				today->tm_sec = 0;
				today->tm_min = 0;
				NextScanPromo = mktime(today);

				while (NextScanPromo <= timer) {
					NextScanPromo = NextScanPromo + 
						(mediabox404_config.promo_timing * 60);
				}
				mediabox404_log_debug("NextScanPromo\n");
			}
		}

		/* "Jingle" scheduling function */
		if (timer >= NextScanJingle) {
			/* it's time to schedule a jingle... */
			if (flag_doublon_jingle == 0) {
				flag_doublon_jingle = 1;
				if (StreamType == MAIN_STREAM) {
					Insert_Jingle(&mysql, &mediabox404_config);
				}

				today = localtime(&timer);
				today->tm_sec = 0;
				today->tm_min = 0;
				NextScanJingle = mktime(today);
				while (NextScanJingle <= timer) {
					NextScanJingle = NextScanJingle + 
						(mediabox404_config.jingle_timing * 60);
				}
				mediabox404_log_debug("NextScanJingle");
			}
		}

		/* "Track" scheduling function */
		if (timer >= NextScanProgramme) {
			// les jingle des prochains programmes sur 24h
			if ( (StreamType == MAIN_STREAM) && (flag_doublon_jingle_prg == 0) ) {
				flag_doublon_jingle_prg = 1;
				mediabox404_log_debug("Insert_Jingle_Prg_24");
				Insert_Jingle_Prg_24(&mysql, &mediabox404_config, NextScanProgramme);
				mediabox404_log_debug("Insert_Jingle_Prg_24 : ok");
			}

			prg_id = 0;
			mediabox404_log_debug("New_Program");
			New_Program(&mysql, &mediabox404_config, date_prg, &prg_id, &data_id, &prg_type, &PlayTime, MetaData, &AutoProgDelay);
			mediabox404_log_debug("New_Program : ok");

			if (prg_id > 0) {
				mediabox404_log_debug("New_Program > 0");

				// reprogrammation automatique
				if (AutoProgDelay > 0) {
					NextAutoProgramme = NextScanProgramme + (AutoProgDelay * 60);
					Copy_Program(&mysql, &mediabox404_config, &NextScanProgramme, &NextAutoProgramme);
				}

				// metadata
				if (strlen(MetaData) > 0) {
					AddMetaData(&mysql, &mediabox404_config, MetaData);
				}

				// copy pile stream --> t_tempstack, delete pile stream
				Copy_streamstack_tempstack(&mysql, &mediabox404_config);

				switch (prg_type) {

					case PL_STREAM :
						mediabox404_log_debug("New Programme PL: %d", prg_id);
						Insert_Log(&mysql, &mediabox404_config, 0, "PL", prg_id, "", "", "", "", "", "", "", "");
						Reset_Last_Playlist(&mysql,&mediabox404_config);
						if (StreamType == RELAY_STREAM) {
							/* this is a relay */
							flag_stop_relay = 1;
						}
						StreamType = PL_STREAM;
						CurrentPlaylist = prg_id;
						break;

					case RELAY_STREAM :
						mediabox404_log_debug("New Programme RELAY: %d",
								prg_id);
						if (StreamType == RELAY_STREAM) {
							/* this is a relay */
							flag_stop_relay = 1;
						}
						StreamType = RELAY_STREAM;
						CurrentPlaylist = prg_id;
						break;

					case REDIF_STREAM :
						mediabox404_log_debug("New Program REDIF: %d", prg_id);
						Insert_Log(&mysql, &mediabox404_config, 0, "REDIF", prg_id, "", "", "", "", "", "", "", "");
						if (StreamType == RELAY_STREAM) {
							/* this is a relay */
							flag_stop_relay = 1;
						}
						StreamType = REDIF_STREAM;
						CurrentPlaylist = prg_id;
						CurrentData = data_id;
						break;
				}
			}

			today = localtime(&timer);
			today->tm_sec = 0;
			today->tm_min = 0;
			NextScanProgramme = mktime(today);
			while (NextScanProgramme <= timer) {
				NextScanProgramme = NextScanProgramme + (mediabox404_config.program_timing * 60);
			}
			mediabox404_log_debug("NextScanProgram");
		}

		// Scan �enement programme WWW
		if (ScanPrg > SCAN_PROG_STACK) { /* if time superior to maximum of time to check the playlist */ 
			//checkList();

			/*ScanPrg = 0;
			  sprintf(query_prgstack, "select * from %s order by S_ID ASC",
			  mediabox404_config.sql.progstack);
			  query_ok = mysql_real_query(&mysql, query_prgstack,
			  (unsigned int) 
			  strlen(query_prgstack));
			  if (query_ok == 0) 
			  {
			  res = mysql_store_result(&mysql);
			  if (mysql_num_rows(res) > 0) 
			  {
			  while ((row = mysql_fetch_row(res))) 
			  {
			  lengths = mysql_fetch_lengths(res);
			  sprintf(CODE,"%.*s", (int) lengths[2], 
			  row[2] ? row[2] : "NULL"); // Command Code LIVE
			  StackPrgID = atoi(row[0]);

			//logfile("Pilot=%s\n",CODE);
			if (strcmp(CODE,"LIVE") == 0) 
			{
			UserPrgID = atoi(row[1]);
			DataPrgID = atoi(row[4]);
			sprintf(ARG[0],"%.*s", (int) lengths[5], row[5] ? row[5] : "NULL"); // URL
			sprintf(ARG[1],"%.*s", (int) lengths[6], row[6] ? row[6] : "NULL"); // Format
			LIVE_Bitrate = atoi(row[7]);// Bitrate [16000,24000,32000,48000,56000,64000,80000,96000,112000,128000,160000,196000,256000]
			LIVE_Frequence = atoi(row[8]); // Fr�uence [16000,22050,24000,32000,44100,48000]
			LIVE_NbCanal = atoi(row[9]);// Nb canaux [1,2] Mono/St��
			LIVE_NbReconnect = atoi(row[10]); // Nb reconnect
			//insert LIVE dans T_StreamStack;
			Insert_Log(&mysql,&mediabox404_config,0,"LIVE",DataPrgID,"","","","","","","","");
			Insert_streamstack_live(&mysql,&mediabox404_config,UserPrgID,DataPrgID,ARG[0],ARG[1],LIVE_Bitrate,LIVE_Frequence,LIVE_NbCanal,LIVE_NbReconnect,0);
			// Insert Jingle
			Insert_Jingle(&mysql,&mediabox404_config);
			}

			if (strcmp(CODE,"RUN") == 0) 
			{
			mediabox404_log_debug("End Relay manuel");
			CurrentPlaylist = 0;
			flag_doublon_jingle_prg = 0;
			StreamType = MAIN_STREAM;
			// copy pile t_tempstack --> stream , delete pile t_tempstack
			Copy_tempstack_streamstack(&mysql,&mediabox404_config);
			do {Insert_Main_PL(&mysql,&mediabox404_config,&nb_control_play);} while (nb_control_play != 0);
			// on arret le LIVE
			Stop_Current_Stream(&mysql,&mediabox404_config);
			EndRelay = 0;
			}
			// Purge ��ement admin ID 
			Delete_ProgStackID (&mysql,&mediabox404_config,StackPrgID);
			wait = 1;
			}
			}

			mysql_free_result(res);
			if (wait == 1) 
			{
			mediabox404_log_debug("Waiting for a Web command %s",
			mediabox404_config.sql.mysql_options);
			wait = 0;
			}
			}*/
		}

		// Scan pile de morceau vide ?
		sprintf(query_streamstack, "SELECT * FROM %s ORDER BY S_ID ASC", mediabox404_config.sql.streamstack);
// 		mediabox404_log_debug( "SCHEDULING MAIN LOOP : Executing query : %s", query_streamstack );

		query_ok = mysql_real_query(&mysql, query_streamstack, (unsigned int)strlen(query_streamstack));
		if (query_ok == 0) {
			MYSQL_RES *res;
			res=mysql_store_result(&mysql);

			if (mysql_num_rows(res) == 0) {
				mediabox404_log_debug( "SCHEDULING MAIN LOOP : There is no result from query (%s).", query_streamstack );
				//logfile("StreamType=%d\n",StreamType);
				// lib�ation des jingle et promo
				flag_doublon_jingle = 0;
				flag_doublon_promo = 0;

				switch (StreamType) {

					case MAIN_STREAM :
						mediabox404_log_debug( "SCHEDULING MAIN LOOP : case MAIN_STREAM" );
						flag_doublon_jingle_prg = 0;
						do {
							Insert_Main_PL(&mysql, &mediabox404_config, &nb_control_play);
						} while (nb_control_play != 0);
						break;

					case RELAY_STREAM :
						mediabox404_log_debug( "SCHEDULING MAIN LOOP : case RELAY_STREAM" );
						if (EndRelay == 0) {
							if (Insert_Relay(&mysql,&mediabox404_config, CurrentPlaylist,PlayTime) == 1) {
								Insert_Log(&mysql, &mediabox404_config, 0, "RELAY", CurrentPlaylist, "", "", "", "", "", "", "", "");
								Insert_Jingle(&mysql, &mediabox404_config);
								Insert_Jingle_end(&mysql, &mediabox404_config, &nb_control_play);
								Stop_Current_Stream(&mysql, &mediabox404_config);
								EndRelay = timer + (PlayTime/1000);
								// pas de jingle ni de promo dans un relay
								flag_doublon_jingle = 1;
								flag_doublon_promo = 1;
							} else {
								mediabox404_log_debug("No Relay");
								CurrentPlaylist = 0;
								flag_doublon_jingle_prg = 0;
								StreamType = MAIN_STREAM;
								// copy pile t_tempstack --> stream , delete pile t_tempstack
								Copy_tempstack_streamstack(&mysql, &mediabox404_config);
								do {
									Insert_Main_PL(&mysql, &mediabox404_config, &nb_control_play);
								} while (nb_control_play != 0);
							}
						} else { // fin plus de stream relay !!
							if (flag_stop_relay == 0) {
								mediabox404_log_error("End Relay deconnect");
								CurrentPlaylist = 0;
								flag_doublon_jingle_prg = 0;
								StreamType = MAIN_STREAM;
								// copy pile t_tempstack --> stream , delete pile t_tempstack
								Copy_tempstack_streamstack(&mysql, &mediabox404_config);
								do {
									Insert_Main_PL(&mysql, &mediabox404_config, &nb_control_play);
								} while (nb_control_play != 0);
								// on arret le LIVE
								Stop_Current_Stream(&mysql,&mediabox404_config);
								EndRelay = 0;
							}
						}
						break;

					case PL_STREAM :
						mediabox404_log_debug( "SCHEDULING MAIN LOOP : case PL_STREAM" );
						if (Insert_PL(&mysql,&mediabox404_config, CurrentPlaylist,&nb_control_play) == 0) {
							// Fin de la pl
							mediabox404_log_debug("End Program");
							CurrentPlaylist = 0;
							flag_doublon_jingle_prg = 0;
							StreamType = MAIN_STREAM;
							// copy pile t_tempstack --> stream , delete pile t_tempstack
							Copy_tempstack_streamstack(&mysql,&mediabox404_config);
							do {
								Insert_Main_PL(&mysql,&mediabox404_config, &nb_control_play);
							} while (nb_control_play != 0);
						} else {
							// pas de jingle ni de promo dans un programme
							flag_doublon_jingle_prg = 1;
							flag_doublon_jingle = 1;
							flag_doublon_promo = 1;
						}
						break;

					case REDIF_STREAM :
						mediabox404_log_debug( "SCHEDULING MAIN LOOP : case REDIF_STREAM" );
						mediabox404_log_debug("REDIF_STREAM : DataID:%d", CurrentData);
						// copy pile t_tempstack --> stream , delete pile t_tempstack
						Copy_tempstack_streamstack(&mysql,&mediabox404_config);
						Insert_streamstack_data_begin(&mysql, &mediabox404_config, CurrentPlaylist, CurrentData, 1, "REDIF");
						// Playlist == 1 Data
						CurrentPlaylist = 0;
						StreamType = MAIN_STREAM;
						break;
				}
			} else {
// 				mediabox404_log_debug( "SCHEDULING MAIN LOOP : There is some results from query. (nothing to do)" );
			}
			mysql_free_result(res);
		}

		// fin d'un relay auto
		if (StreamType == RELAY_STREAM) {
			if ((timer >= EndRelay) && (EndRelay > 0)) {
				// c'est la fin du relay
				if (flag_stop_relay == 0) {
					mediabox404_log_debug("End Relay auto");
					CurrentPlaylist = 0;
					flag_doublon_jingle_prg = 0;
					StreamType = MAIN_STREAM;
					// copy pile t_tempstack --> stream , delete pile t_tempstack
					Copy_tempstack_streamstack(&mysql,&mediabox404_config);
					do {
						Insert_Main_PL(&mysql,&mediabox404_config,
								&nb_control_play);
					} while (nb_control_play != 0);
					// on arret le LIVE
					Stop_Current_Stream(&mysql,&mediabox404_config);
					EndRelay = 0;
				}
			}
		}
		// fin d'un relay par un autre programme
		if (flag_stop_relay == 1) {
			flag_stop_relay = 0;
			// on arret le Relay
			if (StreamType != RELAY_STREAM) {
				Stop_Current_Stream(&mysql,&mediabox404_config);
			}
			EndRelay = 0;
		}

		ScanPrg++;
		sleep(1);
	}

	mysql_close(&mysql);
	mediabox404_log("Exit Programmation Mediabox404_audio");

	return EXIT_SUCCESS;
}
